using System;
using System.Collections;
using System.Linq;
using System.Collections.Generic;
using System.Reflection;

namespace Michigan.Ods.Sdk.Helpers
{
    public class PatchApplier
    {
        public void Execute(object entity, object patch)
        {
            if (entity == null)
                return;

            if (patch == null)
                return;

            var entityType = entity.GetType();
            var patchType = patch.GetType();

            var entitySdkMetaData = new SdkMetaDataWrapper(entity);
            var patchSdkMetaData = new SdkMetaDataWrapper(patch);

            if (entityType != patchType)
                throw new ArgumentException("entity and patch must be the same type.");

            if( entitySdkMetaData.HashCode != patchSdkMetaData.HashCode)
                throw new InvalidOperationException("Entity instance and Patch instance do not have the same natural key.");

            var propertyInfos = patchType.GetProperties(BindingFlags.Public | BindingFlags.Instance);
            
            foreach (var propertyInfo in propertyInfos)
            {
                if (IsSimpleType(propertyInfo.PropertyType))
                {
                    ApplyPatchToEntityIfSet(propertyInfo, entity, patch);
                }
                else if (typeof(IList).IsAssignableFrom(propertyInfo.PropertyType))
                {
                    var entityCollection = (IList)propertyInfo.GetValue(entity);
                    var entityDictionary = CollectionToDictionary(entityCollection);
                    var patchDictionary = CollectionToDictionary(propertyInfo.GetValue(patch) as IEnumerable);

                    foreach (var currentPatchItem in patchDictionary)
                    {
                        if (entityDictionary.ContainsKey(currentPatchItem.Key))
                        {
                            var currentEntityItem = entityDictionary[currentPatchItem.Key];
                            Execute(currentEntityItem, currentPatchItem.Value);
                        }
                        else
                        {
                            if (entityCollection == null)
                            {
                                entityCollection = (IList)propertyInfo.PropertyType.GetConstructor(Type.EmptyTypes).Invoke(null);
                                propertyInfo.SetValue(entity, entityCollection);
                            }
                            entityCollection.Add(currentPatchItem.Value);
                        }
                    }
                }
                else
                {
                    Execute(propertyInfo.GetValue(entity), propertyInfo.GetValue(patch));
                }
            }
        }

        public Dictionary<int, object> CollectionToDictionary(IEnumerable collection)
        {
            if( collection == null)
                return new Dictionary<int, object>();

            var dictionary = new Dictionary<int, object>();
            foreach (var currentObject in collection)
            {
                var sdkMetaData = new SdkMetaDataWrapper(currentObject);
                dictionary.Add(sdkMetaData.HashCode, currentObject);
            }
            return dictionary;
        }

        public void ApplyPatchToEntityIfSet(PropertyInfo propertyInfo, object entity, object patch)
        {
            var patchValue = propertyInfo.GetValue(patch);

            if (patchValue != GetDefaultValue(propertyInfo.PropertyType))
                propertyInfo.SetValue(entity, patchValue);
        }

        public static bool IsSimpleType(Type type)
        {
            return
                type.IsValueType ||
                type.IsPrimitive ||
                new Type[] { 
                typeof(String),
                typeof(Decimal),
                typeof(DateTime),
                typeof(DateTimeOffset),
                typeof(TimeSpan),
                typeof(Guid)
            }.Contains(type) ||
                Convert.GetTypeCode(type) != TypeCode.Object;
        }

        private static object GetDefaultValue(Type t)
        {
            return t.IsValueType ? Activator.CreateInstance(t) : null;
        }
    }

    public class SdkMetaDataWrapper
    {
        private static readonly SdkMetadataCacheLoader SdkMetadataCacheLoader = new SdkMetadataCacheLoader();
        private readonly int _hashCode;

        public SdkMetaDataWrapper(object myObject)
        {
            this.MyObject = myObject;

            var sdkMetaData = SdkMetadataCacheLoader.GetCacheData().FirstOrDefault(x => x.TypeName == myObject.GetType().Name);
            if (sdkMetaData == null)
            {
                this.PrimaryPropertyInfos = new List<PropertyInfo>();
                _hashCode = int.MinValue;
                return;
            }

            var primaryKeyProperties = sdkMetaData.MetadataProperties.Where(x => x.IsPartOfNaturalKey).ToList();

            this.PrimaryPropertyInfos =
                myObject.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance)
                    .Join(primaryKeyProperties, x => x.Name, y => y.PropertyName,
                        (info, propertyInfo) => new { info, propertyInfo }).Select(x => x.info).ToList();

            _hashCode = GetHashCode(0, myObject, this.PrimaryPropertyInfos);
        }

        public int HashCode
        {
            get { return _hashCode; }
        }

        public object MyObject { get; private set; }

        public List<PropertyInfo> PrimaryPropertyInfos { get; private set; }

        public static int GetHashCode(int hashCode, object currentObject, List<PropertyInfo> propertyInfos)
        {
            foreach (var propertyInfo in propertyInfos)
            {
                var propertyInfoValue = propertyInfo.GetValue(currentObject);
                if (propertyInfoValue == null)
                    continue;
                if (PatchApplier.IsSimpleType(propertyInfo.PropertyType))
                {
                    if (propertyInfo.PropertyType == typeof(int))
                        hashCode = (hashCode * 397) ^ (int)propertyInfoValue;
                    else
                        hashCode = (hashCode * 397) ^ propertyInfoValue.GetHashCode();
                }
                else if (typeof(ICollection<>).IsAssignableFrom(propertyInfo.PropertyType) || typeof(IList<>).IsAssignableFrom(propertyInfo.PropertyType) ||
                    typeof(IList).IsAssignableFrom(propertyInfo.PropertyType))
                {

                }
                else
                {
                    hashCode = GetHashCode(hashCode, propertyInfoValue,
                        propertyInfo.PropertyType.GetProperties(BindingFlags.Public | BindingFlags.Instance).ToList());
                }
            }
            return hashCode;
        }
    }
}
